#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 FragPos;
	vec2 uv;
	smooth vec3 normal_world;
	vec3 viewPos;
	vec3 FragPos_;
	vec3 viewPos_;
	vec4 FragPosLightSpace;
	float vis;
} vert;

in struct DirectionLightTangent
{
	vec3 color;
	vec3 direction;
	vec3 directionN;
	bool enabled;
} dirLT;

in struct PointLightTangent {
	vec3 color;
	vec3 position;
	vec3 attenuation;
	bool enabled;
} pointLT;



//uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular
uniform vec3 materialAmbient;
uniform vec3 materialDiffuse;
uniform vec3 materialSpecular;
uniform float specularAlpha;

uniform sampler2D texture_diffuse1;
uniform sampler2D texture_specular1;
uniform sampler2D texture_normal1;

uniform sampler2D shadowMap;
uniform float bokehRadius;
uniform vec3 fogColor;

const float Epsilon = 1e-10;

uniform int enaDiff1;
uniform int enaSpec1;
uniform int enaNorm1;
uniform int drawNormal;
uniform samplerCube skybox;
uniform float gamma;
uniform bool showDepthMap;

uniform float focalDepth;
uniform float focalRange;
uniform sampler2D depthMap;
uniform float nearZ;
uniform float farZ;

layout (location = 0) out vec4 color;
layout (location = 1) out vec4 BrightColor;

float ShadowCalculation(vec4 fragPosLightSpace, float bias)
{
	// perform perspective divide
  vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
	// transform to [0,1] range
	projCoords = projCoords * 0.5 + 0.5;
	// get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
	float closestDepth = texture(shadowMap, projCoords.xy).r;
	// get depth of current fragment from light's perspective
	float currentDepth = projCoords.z;
	// check whether current frag pos is in shadow

	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
	for(int x = -1; x <= 1; ++x)
	{
	    for(int y = -1; y <= 1; ++y)
	    {
	        float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r;
	        shadow += currentDepth - bias > pcfDepth ? 1.0 : 0.0;
	    }
	}
  shadow /= 9.0;
	if(projCoords.z > 1.0)
        shadow = 0.0;
	return shadow;
}

float linearize_Z(float depth , float near , float far){

			float z = depth * 2.0 - 1.0; // back to NDC
	    return (2.0 * near * far) / (far + near - z * (far - near));
}

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 lightColor, vec3 materialDiff, vec3 lightColorSpec, vec3 materialSpec, float alpha, bool attenuate, vec3 attenuation, vec3 specColor, vec3 texColor, vec3 materialAmb) {
	float d = length(l);
	l = normalize(l);

	vec3 r = reflect(-l, n);
	float dnl = dot(l, n);
	float diff = max(0.0, dnl);
	if(alpha <= 0.0) alpha = 1.0;
	float spec = pow(max(0.0, dot(v, r)), alpha);

  vec3 ambient = materialDiff * texColor * lightColor;
	vec3 diffuse = materialDiff * texColor * lightColor * diff;
	vec3 specular = materialSpec * specColor * lightColorSpec * spec;

	float bias = max(0.05 * (1.0 - dnl), 0.005);
	float shadow = ShadowCalculation(vert.FragPosLightSpace, bias);
	return (ambient + (1.0 - shadow) * (diffuse + specular)) ;
}

vec3 RGBtoHSV(in vec3 RGB)
{
    vec4  P   = (RGB.g < RGB.b) ? vec4(RGB.bg, -1.0, 2.0/3.0) : vec4(RGB.gb, 0.0, -1.0/3.0);
    vec4  Q   = (RGB.r < P.x) ? vec4(P.xyw, RGB.r) : vec4(RGB.r, P.yzx);
    float C   = Q.x - min(Q.w, Q.y);
    float H   = abs((Q.w - Q.y) / (6.0 * C + Epsilon) + Q.z);
    vec3  HCV = vec3(H, C, Q.x);
    float S   = HCV.y / (HCV.z + Epsilon);
    return vec3(HCV.x, S, HCV.z);
}

vec3 HSVtoRGB(in vec3 HSV)
{
    float H   = HSV.x;
    float R   = abs(H * 6.0 - 3.0) - 1.0;
    float G   = 2.0 - abs(H * 6.0 - 2.0);
    float B   = 2.0 - abs(H * 6.0 - 4.0);
    vec3  RGB = clamp( vec3(R,G,B), 0.0, 1.0 );
    return ((RGB - 1.0) * HSV.y + 1.0) * HSV.z;
}

void main() {

	vec3 normal = vert.normal_world;
	if(enaNorm1 == 1)
		if( textureSize( texture_normal1, 0).x > 0)
		{
			normal = texture(texture_normal1, vert.uv).rgb;
			normal = normalize(normal * 2.0 - 1.0);
		}


	vec4 texxColor = vec4(1,1,1, 1);
	vec4 speccColor = vec4(1,1,1, 1);

	if(enaDiff1 == 1)
		if( textureSize( texture_diffuse1, 0).x > 0)
			texxColor = texture(texture_diffuse1, vert.uv);

  //discard for alpha
	if(texxColor.a < 0.1)
	   discard;

	if(enaSpec1 == 1)
		if( textureSize( texture_specular1, 0).x > 0)
			speccColor = texture(texture_specular1, vert.uv);


	vec3 specColor = speccColor.rgb;
	vec3 texColor = texxColor.rgb;

	vec3 v = normalize(vert.viewPos - vert.FragPos);

	// add directional light contribution
	color.rgb += phong(normal, -dirLT.direction, v, dirLT.color , materialDiffuse , dirLT.color, materialSpecular , specularAlpha, false, vec3(1), specColor, texColor, materialAmbient);


	///add Skybox reflection
	vec3 R = reflect(v, normalize(normal));
	color = mix(color, vec4(texture(skybox, R).rgb, 1.0), 0.005);

	//de-saturate colors
  //vec3 col_hsv = RGBtoHSV(color.rgb);
	//col_hsv.y *= 0.8;
	//color.rgb = vec3(HSVtoRGB(col_hsv));

  //calculating the depth and the Circle of Confusion (COC) for depth of field
	float depth = linearize_Z(gl_FragCoord.z, nearZ, farZ);
	float coc = abs(focalDepth + depth) / focalRange;
	coc = clamp(coc, -1.0, 1.0);

	//store the coc for this fragment in the alpha channel
	color.a = coc;

	//filter brightness
	float brightness = dot(color.rgb, vec3(0.2126, 0.7152, 0.0722));
	if (brightness > 0.8 )
	{
		vec3 op = clamp(color.rgb, vec3(0), vec3(255));
		BrightColor = vec4(op, 	color.a);
	}else
		BrightColor = vec4(0.0, 0.0, 0.0, 	color.a);

  //draws the COC/depth map for DOF when enabled
	if(showDepthMap)
	{
		color.rgb = vec3(coc);
	}

  //draws the normals if enabled
	if(drawNormal == 1)
	{
		color = vec4(normal, 1);
	}
}
