#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 FragPos;
	vec2 uv;
	smooth vec3 normal_world;
	vec3 viewPos;
	vec3 FragPos_;
	vec3 viewPos_;
	vec4 FragPosLightSpace;
	float vis;
} vert;

in struct DirectionLightTangent
{
	vec3 color;
	vec3 direction;
	vec3 directionN;
	bool enabled;
} dirLT;

in struct PointLightTangent {
	vec3 color;
	vec3 position;
	vec3 attenuation;
	bool enabled;
} pointLT;



//uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular
uniform vec3 materialAmbient;
uniform vec3 materialDiffuse;
uniform vec3 materialSpecular;
uniform float specularAlpha;

uniform sampler2D texture_diffuse1;
uniform sampler2D texture_specular1;
uniform sampler2D texture_normal1;

uniform sampler2D shadowMap;
uniform float bokehRadius;
uniform vec3 fogColor;

const float Epsilon = 1e-10;

uniform int enaDiff1;
uniform int enaSpec1;
uniform int enaNorm1;
uniform int drawNormal;
uniform samplerCube skybox;
uniform float gamma;
uniform bool showDepthMap;

uniform float focalDepth;
uniform float focalRange;
uniform sampler2D depthMap;
uniform float nearZ;
uniform float farZ;

layout (location = 0) out vec4 color;
layout (location = 1) out vec4 BrightColor;

uniform vec3  albedo;
uniform float metallic = 0.1;
uniform float roughness = 0.9;
uniform float ao = 5.0;
const float PI = 3.14159265359;
bool useTexture = false;

float DistributionGGX(vec3 N, vec3 H, float roughness);
float GeometrySchlickGGX(float NdotV, float roughness);
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness);
vec3 fresnelSchlick(float cosTheta, vec3 F0);

float ShadowCalculation(vec4 fragPosLightSpace, float bias)
{
	// perform perspective divide
  vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
	// transform to [0,1] range
	projCoords = projCoords * 0.5 + 0.5;
	// get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
	float closestDepth = texture(shadowMap, projCoords.xy).r;
	// get depth of current fragment from light's perspective
	float currentDepth = projCoords.z;
	// check whether current frag pos is in shadow

	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
	for(int x = -1; x <= 1; ++x)
	{
	    for(int y = -1; y <= 1; ++y)
	    {
	        float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r;
	        shadow += currentDepth - bias > pcfDepth ? 1.0 : 0.0;
	    }
	}
  shadow /= 9.0;
	if(projCoords.z > 1.0)
        shadow = 0.0;
	return shadow;
}

float linearize_Z(float depth , float near , float far){

			float z = depth * 2.0 - 1.0; // back to NDC
	    return (2.0 * near * far) / (far + near - z * (far - near));
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
	if(cosTheta>1.0)
	cosTheta=1.0;
	float p=pow(1.0 - cosTheta, 5.0);
	return F0 + (1.0 - F0) * p;
}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;

    return num / denom;
}
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);

    return ggx1 * ggx2;
}



vec3 RGBtoHSV(in vec3 RGB)
{
    vec4  P   = (RGB.g < RGB.b) ? vec4(RGB.bg, -1.0, 2.0/3.0) : vec4(RGB.gb, 0.0, -1.0/3.0);
    vec4  Q   = (RGB.r < P.x) ? vec4(P.xyw, RGB.r) : vec4(RGB.r, P.yzx);
    float C   = Q.x - min(Q.w, Q.y);
    float H   = abs((Q.w - Q.y) / (6.0 * C + Epsilon) + Q.z);
    vec3  HCV = vec3(H, C, Q.x);
    float S   = HCV.y / (HCV.z + Epsilon);
    return vec3(HCV.x, S, HCV.z);
}

vec3 HSVtoRGB(in vec3 HSV)
{
    float H   = HSV.x;
    float R   = abs(H * 6.0 - 3.0) - 1.0;
    float G   = 2.0 - abs(H * 6.0 - 2.0);
    float B   = 2.0 - abs(H * 6.0 - 4.0);
    vec3  RGB = clamp( vec3(R,G,B), 0.0, 1.0 );
    return ((RGB - 1.0) * HSV.y + 1.0) * HSV.z;
}

void main() {

	vec3 normal = vert.normal_world;
	if(enaNorm1 == 1)
		if( textureSize( texture_normal1, 0).x > 0)
		{
			normal = texture(texture_normal1, vert.uv).rgb;
			normal = normalize(normal * 2.0 - 1.0);
		}


	vec3 clr = albedo;

	if(enaDiff1 == 1)
		if( textureSize( texture_diffuse1, 0).x > 0)
		{
			vec4 col = texture(texture_diffuse1, vert.uv);
			if(col.a < 0.1)
				 discard;
			clr = col.rgb;
			useTexture = true;
		}
	//discard for alpha


	vec3 V = normalize(- vert.FragPos_);
	vec3 N = normal;

	vec3 F0 = vec3(0.04);
	F0 = mix(F0, clr, metallic);

	// reflectance equation
	vec3 Lo = vec3(0.0);

	// calculate per-light radiance
	vec3 L = normalize(-dirLT.direction);//normalize(lightPositions[i] - WorldPos);
	vec3 H = normalize(V + L);
	//float distance    = length(lightPositions[i] - WorldPos);
	//float attenuation = 1.0 / (distance * distance);
	vec3 radiance = dirLT.color; //lightColors[i] * attenuation;

	// cook-torrance brdf
	float NDF = DistributionGGX(N, H, roughness);
	float G   = GeometrySmith(N, V, L, roughness);
	vec3 F    = fresnelSchlick(max(dot(H, V), 0.0), F0);

	vec3 kS = F;
	vec3 kD = vec3(1.0) - kS;
	kD *= 1.0 - metallic;

	vec3 numerator    = NDF * G * F;
	float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0);
	vec3 specular     = numerator / max(denominator, 0.001);

	// add to outgoing radiance Lo
	float NdotL = max(dot(N, L), 0.0);
	Lo += (kD * clr / PI + specular) * radiance * NdotL;

	float dnl = dot(normalize(-dirLT.direction), N);
	float bias = max(0.05 * (1.0 - dnl), 0.005);
	float shadow = ShadowCalculation(vert.FragPosLightSpace, bias);
	vec3 clrDown = vec3(0.03);
	if(useTexture)
		clrDown = vec3(0.1);
	vec3 ambient = (clr + (1.0 - shadow)) * (clrDown * clr * ao);
	color.rgb =  ambient + Lo;

	///add Skybox reflection
	vec3 R = reflect(V, normalize(normal));
	color = mix(color, vec4(texture(skybox, R).rgb, 1.0), 0.005);

	//saturate colors
//  vec3 col_hsv = RGBtoHSV(color.rgb);
//	col_hsv.y *= 1.2;
//	color.rgb = vec3(HSVtoRGB(col_hsv));

  //calculating the depth and the Circle of Confusion (COC) for depth of field
	float depth = linearize_Z(gl_FragCoord.z, nearZ, farZ);
	float coc = abs(focalDepth + depth) / focalRange;

	coc = clamp(coc, -1.0, 1.0);

	//store the coc for this fragment in the alpha channel
	color.a = coc;

	//filter brightness
	float brightness = dot(color.rgb, vec3(0.2126, 0.7152, 0.0722));
	if (brightness > 1.0 )
	{
		vec3 op = clamp(color.rgb, vec3(0), vec3(255));
		BrightColor = vec4(op, 	color.a);
	}else
		BrightColor = vec4(0.0, 0.0, 0.0, 	color.a);

  //draws the COC/depth map for DOF when enabled
	if(showDepthMap)
	{
		color.rgb = vec3(coc);
	}

  //draws the normals if enabled
	if(drawNormal == 1)
	{
		color = vec4(normal, 1);
	}
}
